"""Local 'database' of runs class."""

import json
import os

from curifactory import utils


class ManagerStore:
    """Manages the mini database of metadata on previous experiment runs. This is how we
    keep track of experiment run numbers etc. A metadata block for each run is stored in
    the manager cache path under :code:`store.json`.

    Note that the metadata blocks we keep track of for each run follows the following example:

    .. code-block:: json

        {
            "reference": "example_experiment_1_2021-06-15-T100003",
            "experiment_name": "example_experiment",
            "run_number": 1,
            "timestamp": "2021-06-15-T100003",
            "commit": "",
            "params_files": ["example_params"],
            "args": { "example_params": [ [ "test_params", "44b5e428e7165975a3e4f0d1674dbe5f" ] ] },
            "full_store": false,
            "status": "complete",
            "cli": "experiment example_experiment -p example_params",
            "hostname": "mycomputer",
            "notes": ""
        }

    Args:
        manager_cache_path (str): The path to the directory to keep the :code:`store.json`.
    """

    def __init__(self, manager_cache_path: str):
        self.runs = []
        """The list of metadata blocks for each run."""
        self.path = manager_cache_path
        """The location to store the :code:`store.json`."""

        if self.path[-1] != "/":
            self.path += "/"

        self.path += "store.json"

        self.load()

    def load(self):
        """Load the current experiment database from :code:`sore.json` into :code:`self.runs`."""
        if os.path.exists(self.path):
            with open(self.path, "r") as infile:
                self.runs = json.load(infile)

    def save(self):
        """Save the current database in :code:`self.runs` into the :code:`store.json` file."""
        with open(self.path, "w") as outfile:
            json.dump(self.runs, outfile, indent=4)

    def get_experiment_runs(self, experiment_name: str):
        """Get all the runs associated with the specified experiment name from the database.

        Args:
            experiment_name (str): The experiment name to get all run metadata for.

        Returns:
            A list of all dictionaries (metadata blocks) that have the requested experiment name.
        """
        experiment_runs = []
        for run in self.runs:
            if run["experiment_name"] == experiment_name:
                experiment_runs.append(run)
        return experiment_runs

    def get_run(self, ref_name: str):
        """Get the metadata block for the run with the specified reference name.

        Args:
            ref_name (str): The run reference name, following the [experiment_name]_[run_number]_[timestamp] format.

        Returns:
            A dictionary (metadata block) for the run with the requested reference name, and the
            index of the run within the total list of runs.
        """
        for index, run in enumerate(self.runs):
            if run["reference"] == ref_name:
                return run, index
        return None, -1

    def add_run(self, mngr):
        """Add a new metadata block to the store for the passed :code:`ArtifactManager` instance.

        Note that this automatically calls the :code:`save()` function.

        Args:
            mngr (ArtifactManager): The manager to grab run metadata from.

        Returns:
            The newly created dictionary (metadata block) for the current manager's run.
        """
        prev_runs = self.get_experiment_runs(mngr.experiment_name)
        if len(prev_runs) == 0:
            mngr.experiment_run_number = 1
        else:
            mngr.experiment_run_number = prev_runs[-1]["run_number"] + 1
        mngr.git_commit_hash = utils.get_current_commit()

        # create the metadata block
        run = {
            "reference": mngr.get_reference_name(),
            "experiment_name": mngr.experiment_name,
            "run_number": mngr.experiment_run_number,
            "timestamp": mngr.get_str_timestamp(),
            "commit": mngr.git_commit_hash,
            "params_files": mngr.experiment_args_file_list,
            "args": mngr.experiment_args,
            "full_store": mngr.store_entire_run,
            "status": "incomplete",
            "cli": mngr.run_line,
            "hostname": mngr.hostname,
            "notes": mngr.notes,
        }

        # sanitize reproduction cli command
        if mngr.store_entire_run:
            run = self._get_reproduction_line(mngr, run)

        self.runs.append(run)

        self.save()
        return run

    # NOTE: we have to call this both from add_run and update_run because manager stores itself on init, but if someone _later_ sets store_full (maybe in a live run) we need to be able to handle this being added to the run_info
    def _get_reproduction_line(self, mngr, run):
        sanitized_run_line = mngr.run_line
        if "--overwrite " in sanitized_run_line:
            sanitized_run_line = sanitized_run_line.replace("--overwrite ", "")
        if sanitized_run_line.endswith("--overwrite"):
            sanitized_run_line = sanitized_run_line[:-12]
        sanitized_run_line = sanitized_run_line.replace("--store-full ", "")
        if sanitized_run_line.endswith("--store-full"):
            sanitized_run_line = sanitized_run_line[:-13]

        store_entire_run_path = os.path.join(mngr.runs_path, mngr.get_reference_name())
        mngr.reproduction_line = (
            f"{sanitized_run_line} --cache {store_entire_run_path} --dry-cache"
        )

        run["reproduce"] = mngr.reproduction_line
        return run

    def update_run(self, mngr):
        """Updates the metadata in the database for the run associated with the passed :code:`ArtifactManager`.

        This is currently just used to update the status and include any error messages if relevant, when an experiment
        finishes running.

        Note that this automatically calls the :code:`save()` function.

        Args:
            mngr (ArtifactManager): The manager to grab run metadata from.

        Returns:
            The updated dictionary (metadata block) for the run. It returns None if the experiment isn't
            found in the database.
        """
        run_info, index = self.get_run(mngr.get_reference_name())
        if index == -1:
            # TODO error?
            return None

        run_info["status"] = mngr.status
        if mngr.status == "error":
            run_info["error"] = mngr.error
        run_info["params_files"] = mngr.experiment_args_file_list
        run_info["args"] = mngr.experiment_args

        if mngr.store_entire_run:
            run_info = self._get_reproduction_line(mngr, run_info)

        self.runs[index] = run_info

        self.save()
        return run_info



# flake8: noqa
from curifactory.staging import (
    stage,
    aggregate,
    InputSignatureError,
    OutputSignatureError,
    EmptyCachersError,
    CachersMismatchError,
)
from curifactory.manager import ArtifactManager
from curifactory.record import Record
from curifactory.procedure import Procedure
from curifactory.args import ExperimentArgs
from curifactory.caching import Lazy

__version__ = "0.6.3"



import curifactory as cf
from curifactory.reporting import JsonReporter
from curifactory.caching import JsonCacher
import os


def test_reportables_cached(configured_test_manager):
    """Re-running a stage with cached outputs and reportables should still reload the old reportables."""

    @cf.stage(None, ["stuff"], [JsonCacher])
    def stage_with_reportables(record):
        dictionary = {"thing1": "testing"}
        record.report(JsonReporter(dictionary))

        with open(os.path.join(record.manager.cache_path, "stage_ran"), "w") as outfile:
            outfile.write("\n")

        return dictionary

    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")

    r0 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    stage_with_reportables(r0)

    assert len(configured_test_manager.reportables) == 1
    assert os.path.exists(ran_path)
    os.remove(ran_path)

    r1 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    stage_with_reportables(r1)

    assert len(configured_test_manager.reportables) == 2
    assert not os.path.exists(ran_path)



"""Classes for various caching strategies, known as "cachers"

This is handled through a base :code:`Cacheable` class, and each "strategy"
cacher class extends it.
"""

import json
import logging
import os
import pandas as pd
import pickle
from typing import Union, List


class Lazy:
    """A class to indicate a stage output as a lazy-cache object - curifactory will
    attempt to keep this out of memory as much as possible, immediately caching and deleting,
    and loading back into memeory only when needed."""

    def __init__(self, name: str):
        self.name = name
        self.cacher: Cacheable = None

    def load(self):
        return self.cacher.load()

    def __str__(self):
        return self.name


class Cacheable:
    """The base caching class, any caching strategy should extend this.

    Args:
        extension (str): The filetype extension to add at the end of the path.
        path_override (str): Use a specific path for the cacheable, rather than
            automatically setting it.
    """

    def __init__(self, extension: str, path_override=None):
        self.extension = extension
        """str: The filetype extension to add at the end of the path."""
        self.path = ""
        """str: The path of the cacheable to write out."""
        self.path_override = path_override
        """str: Use a specific path for the cacheable, rather than automatically
            setting it."""
        if self.path_override is not None:
            self.path = self.path_override
        self.record = None
        """Record: the current record this cacheable is caching under. This can be used to get a copy of the current args instance."""

    def set_path(self, path: str):
        """Changes the :code:`path` to the passed value."""
        self.path = path + self.extension
        return self.path

    def check(self) -> bool:
        """Check to see if this cacheable needs to be written or not.

        Note:
            This function will always return False if the args are :code:`None`.

        Returns:
            True if we find the cached file and the current :code:`Args`
            don't specify to overwrite, otherwise False.
        """
        logging.debug("Searching for cached file at %s...", self.path)
        if os.path.exists(self.path):
            if (
                self.record is not None
                and self.record.args is not None
                and not self.record.args.overwrite
            ):
                logging.info("Cached object '%s' found", self.path)
                return True
            else:
                logging.debug("Object found, but overwrite specified in args")
                return False
        logging.debug("Cached file not found")
        return False

    def load(self):
        """Load the cacheable from disk.

        Note:
            Any subclass is **required** to implement this.
        """
        raise NotImplementedError(
            "Cacheable class does not have a load function implemented"
        )

    def save(self, obj):
        """Save the passed object to disk.

        Note:
            Any subclass is **required** to implement this.
        """
        raise NotImplementedError(
            "Cacheable class does not have a save function implemented"
        )


class JsonCacher(Cacheable):
    """Dumps an object to indented JSON."""

    def __init__(self, path_override=None):
        super().__init__(".json", path_override=path_override)

    def load(self):
        with open(self.path, "r") as infile:
            obj = json.load(infile)
        return obj

    def save(self, obj):
        with open(self.path, "w") as outfile:
            json.dump(obj, outfile, indent=4, default=lambda x: str(x))


class PickleCacher(Cacheable):
    """Dumps an object to a pickle file."""

    def __init__(self, path_override=None):
        super().__init__(".pkl.gz", path_override=path_override)

    def load(self):
        with open(self.path, "rb") as infile:
            obj = pickle.load(infile)
        return obj

    def save(self, obj):
        with open(self.path, "wb") as outfile:
            pickle.dump(obj, outfile)


class PandasJsonCacher(Cacheable):
    """Saves a pandas dataframe to JSON."""

    def __init__(self, path_override=None):
        super().__init__(".json", path_override=path_override)

    def load(self):
        return pd.read_json(self.path)

    def save(self, obj):
        obj.to_json(self.path)


class PandasCsvCacher(Cacheable):
    """Saves a pandas dataframe to CSV."""

    def __init__(self, path_override=None):
        super().__init__(".csv", path_override=path_override)

    def load(self):
        return pd.read_csv(self.path)

    def save(self, obj):
        obj.to_csv(self.path)


class FileReferenceCacher(Cacheable):
    """Saves a file path or list of file paths generated from a stage as a json file.
    The :code:`check` function will check existence of all file paths.

    This is useful for instances where there may be a large number of files stored or
    generated to disk, as it would be unwieldy to return them all (or infeasible to keep them
    in memory) directly from the stage. When this cacher is checked for pre-existing,
    it tries to load the json file storing the filenames, and then checks for the existence of
    each path in that json file. If all of them exist, it will short-circuit computation.

    Using this cacher does mean the user is in charge of loading/saving the file paths correctly,
    but in some cases that may be desirable.

    This can also be used for storing a reference to a single file outside the normal cache.

    When combined with the :code:`get_dir` call on the record, you can create a cached directory of
    files similarly to a regular cacher and simply keep a reference to them as part of the actual
    cacher process.

    Example:

        .. code-block:: python

            @stage(inputs=None, outputs=["many_text_files"], cachers=[FileReferenceCacher])
            def output_text_files(record):
                file_path = record.get_dir("my_files")
                my_file_list = [os.path.join(file_path, f"my_file_{num}") for num in range(20)]

                for file in my_file_list:
                    with open(file, 'w') as outfile:
                        outfile.write("test")

                return my_file_list
    """

    def __init__(self, path_override=None):
        super().__init__(".json", path_override=path_override)

    def check(self) -> bool:
        # check the file list file exists
        if not super().check():
            return False

        # load the file list and check each file
        # NOTE: we don't need to re-check args overwrite because that
        # would already have applied in the super check
        with open(self.path, "r") as infile:
            files = json.load(infile)

        if type(files) == list:
            for file in files:
                logging.debug("Checking from file list: '%s'" % file)
                if not os.path.exists(file):
                    return False
        else:
            if not os.path.exists(files):
                return False

        return True

    def load(self) -> Union[List[str], str]:
        with open(self.path, "r") as infile:
            files = json.load(infile)
        return files

    def save(self, files: Union[List[str], str]):
        with open(self.path, "w") as outfile:
            json.dump(files, outfile, indent=4)



"""Contains relevant classes for records, objects that track a particular state
through some set of stages."""

import copy
import logging
import os

from curifactory.caching import Lazy
from curifactory.reporting import Reportable


class CacheAwareDict(dict):
    """A normal dictionary that will return resolved versions of Lazy objects when accessed."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.resolve = True
        """A flag for enabling/disabling auto-resolve. If this is False, this works just
        like a normal dictionary. This is necessary in the input checking part of a normal stage, as
        we make state access before determining if stage execution is required or not."""

    def __getitem__(self, key):
        item = super().__getitem__(key)
        if type(item) == Lazy and self.resolve:
            logging.debug("Auto-resolving lazy object '%s'..." % key)
            return item.cacher.load()
        else:
            return item


# TODO: make printing this dump the state out as a string
# TODO: handle if args is None
class Record:
    """A single persistent state that's passed between stages in a single "experiment line".

    Args:
        manager (ArtifactManager): The artifact manager this record is associated with.
        args: The :code:`ExperimentArgs` instance to apply to any stages this record is run through.
        hide (bool): If :code:`True`, don't add this record to the artifact manager.
    """

    def __init__(self, manager, args, hide=False):
        # TODO: would be nice to be able to initialize a record with a pre-defined state dictionary
        self.manager = manager
        """The :code:`ArtifactManager` associated with this record."""
        self.args = args
        """The :code:`ExperimentArgs` to apply to any stages this record is passed through."""
        # self.state = {}
        self.state = CacheAwareDict()
        """The dictionary of all variables created by stages this record is passed through. (AKA 'Artifacts')
        All :code:`inputs` from stage decorators are pulled from this dictionary, and all
        :code:`outputs` are stored here."""
        self.state_artifact_reps = {}
        """Dictionary mimicking state that keeps an :code:`ArtifactRepresentation` associated with
        each variable stored in :code:`self.state`."""
        self.output = None
        """The returned value from the last stage that was run with this record."""
        self.stages = []
        """The list of stage names that this record has run through so far."""
        self.stage_inputs = []
        """A list of lists per stage with the state inputs that stage requested."""
        self.stage_outputs = []
        """A list of lists per stage with the state outputs that stage produced."""
        self.input_records = []
        """A list of any records used as input to this one. This mostly only occurs when aggregate
        stages are run."""

        if not hide:
            self.manager.records.append(self)

    def report(self, reportable: Reportable):
        """Add a reportable associated with this record, this will get added to the experiment run
        output report.

        Args:
            reportable (Reportable): The reportable to render on the final experiment report.
        """
        reportable.record = self
        reportable.stage = self.stages[-1]

        if reportable.record.args is not None:
            name = f"{reportable.record.args.name}_{reportable.stage}_"
        else:
            name = f"(Aggregate)_{reportable.stage}_"
        if reportable.name is None:
            name += str(len(self.manager.reportables))
        else:
            name += reportable.name
        reportable.name = name

        self.manager.reportables.append(reportable)

    def make_copy(self, args=None, add_to_manager=True):
        """Make a new record that has a deep-copied version of the current state.

        This is useful for a long running procedure that creates a common dataset for
        many other stages, so that it can be replicated across multiple argsets without
        having to recompute for each argset.

        Note that state is really the only thing transferred to the new record, the stage and
        inputs/outputs lists will be empty.

        Also note that the current record will be added to the :code:`input_records` of the new
        record, since it may draw on data in its state.

        Args:
            args: The new :code:`ExperimentArgs` argset to apply to the new record. Leave as None
                to retain the same args as the current record.
            add_to_manager: Whether to automatically add this record to the current manager or not.
        """
        if args is None:
            args = self.args
        new_record = Record(self.manager, args)
        new_record.input_records = [self]
        new_record.state = copy.deepcopy(self.state)
        # TODO: (02/02/2022) state without state artifact reps might cause issues
        # new_record.state_artifact_reps = self.state_artifact_reps

        if add_to_manager:
            self.manager.records.append(new_record)

        return new_record

    def get_path(self, obj_name: str) -> str:
        """Return an args-appropriate cache path with passed object name.

        This should be equivalent to what a cacher for a stage should get. Note that this
        is calling the manager's get_path, which will include the stage name. If calling
        this outside of a stage, it will include whatever stage was last run.

        Warning:
            This path does not yet have a mechanism to be auto-stored in the output folder for --store-full runs. This functionality will be implemented later.

        Args:
            obj_name (str): the name to associate with the object as the last part of the filename.
        """
        # TODO: (02/02/2022) at some point this should keep track of these paths so that store-full can automatically grab them too.
        return self.manager.get_path(obj_name=obj_name, record=self)

    def get_dir(self, dir_name_suffix: str) -> str:
        """Returns an args-appropriate cache path with the passed name, (similar to get_path) and creates it as a directory.

        Warning:
            This path does not yet have a mechanism to be auto-stored in the output folder for --store-full runs. This functionality will be implemented later.

        Args:
            dir_name_suffix (str): the name to add as a suffix to the created directory name.
        """

        # TODO: (02/02/2022) at some point this should keep track of these paths so that store-full can automatically grab them too.
        dir_path = self.manager.get_path(obj_name=dir_name_suffix, record=self)

        os.makedirs(dir_path, exist_ok=True)
        return dir_path


class ArtifactRepresentation:
    """This is a shorthand string representation for an artifact stored in a record as well
    as output cache info. This is what gets displayed in the detailed experiment map in reports.

    This will try to include helpful information in the string representation, such as pandas/
    numpy shapes, or lengths where applicable.

    Args:
        record (Record): The record this artifact is stored in.
        name (str): The name of the artifact.
        artifact: The artifact itself.
    """

    def __init__(self, record, name, artifact):
        self.init_record = record
        self.name = name
        self.string = f"({type(artifact).__name__}) {str(artifact)[:20]}"
        if len(str(artifact)) > 20:
            self.string += "..."
        if hasattr(artifact, "shape"):
            shape = None
            if callable(getattr(artifact, "shape", None)):
                shape = artifact.shape()
            else:
                shape = artifact.shape
            self.string += f" shape: {shape}"
        elif hasattr(artifact, "__len__"):
            self.string += f" len: {len(artifact)}"

        self.file = "no file"

    def html_safe(self):
        """Removes special characters that'd break html, except it doesn't actually display
        correctly.

        (I think this is an issue with how graphviz renders outputs?)"""
        clean = (
            self.string.replace("<", "&lt;")
            .replace(">", "&gt;")
            .replace("&", "&amp;")
            .replace('"', "&quot;")
        )

        return clean



""" Helper and utility functions for the library. """

from copy import deepcopy
from dataclasses import asdict
import hashlib
import json
import logging
import os
import platform
import shutil
import subprocess
import sys
from typing import Dict

TIMESTAMP_FORMAT = "%Y-%m-%d-T%H%M%S"
"""The datetime format string used for timestamps in experiment run reference names."""
CONFIGURATION_FILE = "curifactory_config.json"
"""The expected configuration filename."""

EDITORS = ["vim", "nvim", "emacs", "nano", "vi"]
"""The list of possible editor exec names to test for getting a valid text editor."""


def get_configuration() -> Dict[str, str]:
    """Load the configuration file if available, with defaults for any
    keys not found. The config file should be "curifactory_config.json"
    in the project root.

    The defaults are:

    .. code-block:: json

        {
            "experiments_module_name": "experiments",
            "params_module_name": "params",
            "manager_cache_path": "data/",
            "cache_path": "data/cache",
            "runs_path": "data/runs",
            "logs_path": "logs/",
            "notebooks_path": "notebooks/",
            "reports_path": "reports/",
            "report_css_path": "reports/style.css",
        }

    Returns:
        the dictionary of configuration keys/values.
    """

    config_defaults = {
        "experiments_module_name": "experiments",
        "params_module_name": "params",
        "manager_cache_path": "data/",
        "cache_path": "data/cache",
        "runs_path": "data/runs",
        "logs_path": "logs/",
        "notebooks_path": "notebooks/",
        "reports_path": "reports/",
        "report_css_path": "reports/style.css",
    }

    # try to find configuration file in this dir or parent dirs (
    search_depth = 3
    prefix = ""
    while not os.path.exists(f"{prefix}{CONFIGURATION_FILE}") and search_depth > 0:
        prefix += "../"
        search_depth -= 1

    # get config file
    if os.path.exists(f"{prefix}{CONFIGURATION_FILE}"):
        with open(f"{prefix}{CONFIGURATION_FILE}", "r") as infile:
            config = json.load(infile)

        # in case of any values that don't exist in explicit config
        for key in config_defaults:
            if key not in config:
                config[key] = config_defaults[key]

        # update paths if in subdir
        # NOTE: this doesn't update module names, because that requires actual code changes. Ideally,
        # this is only intended for things like notebooks to be able to use regular paths for live
        # curifactory usage, rather than actual experiment runs. For that it's on the user to correctly
        # ensure their current directory is set.
        for key in config:
            if key.endswith("_path"):
                config[key] = f"{prefix}{config[key]}"
    else:
        # defaults
        config = config_defaults

    return config


def get_editor() -> str:
    """Returns a text editor to use for richer text entry, such as in providing experiment notes."""

    def first_available_editor():
        for editor_name in EDITORS:
            if shutil.which(editor_name) is not None:
                return editor_name
        return None

    editor = None
    if "EDITOR" in os.environ:
        editor = os.getenv("EDITOR")
    elif first_available_editor() is not None:
        editor = first_available_editor()
    elif os.name == "nt":
        editor = "notepad"

    if editor is None:
        raise RuntimeError("No valid text editor could be found.")
    return editor


def human_readable_mem_usage(byte_count: int) -> str:
    """Takes the given byte count and returns a nicely formatted string that includes the suffix (K/M/GB).

    Args:
        byte_count (int): The number of bytes to convert into KB/MB/GB.
    """

    negative = False
    if byte_count < 0:
        negative = True
        byte_count *= -1

    suffix = "B"
    if byte_count > 10 ** 9:
        suffix = "GB"
        byte_count /= 10 ** 9
    elif byte_count > 10 ** 6:
        suffix = "MB"
        byte_count /= 10 ** 6
    elif byte_count > 10 ** 3:
        suffix = "KB"
        byte_count /= 10 ** 3

    if negative:
        return "-%.2f%s" % (byte_count, suffix)
    return "%.2f%s" % (byte_count, suffix)


def human_readable_time(seconds: float) -> str:
    """Takes the given time in seconds and returns a nicely formatted string that includes the suffix.

    Args:
        seconds (float): The time in seconds to convert.
    """

    converted = seconds
    suffix = "s"

    # .1 = 100ms
    # .1ms = 100us = .0001
    # .1us = 100ns = .0000001

    if seconds > 60 * 60:
        suffix = "h"
        converted /= 60 * 60
    elif seconds > 60:
        suffix = "m"
        converted /= 60
    elif seconds < 0.0000001:
        suffix = "ns"
        converted *= 10 ** 9
    elif seconds < 0.0001:
        suffix = "us"
        converted *= 10 ** 6
    elif seconds < 0.1:
        suffix = "ms"
        converted *= 10 ** 3

    return "%.2f%s" % (converted, suffix)


def get_command_output(cmd, silent=False) -> str:
    """Runs the command passed and returns the full string output of the command
    (minus the final newline).

    Args:
        cmd: Either a string command or array of strings, as one would pass to
            :code:`subprocess.run()`
    """
    try:
        cmd_return = subprocess.run(cmd, stdout=subprocess.PIPE)
    except:  # noqa: E722 -- TODO: we should actually handle this for the below note
        # NOTE - seems like we get filenotfound exceptions if it's an invalid command?
        # (e.g. on windows calling git when git isn't found from the env?)
        if not silent:
            logging.warning("Unable to run command '%s'" % cmd)
        return ""
    if cmd_return.returncode == 0:
        output = cmd_return.stdout.decode("utf-8")
        return output[:-2]
    return ""


def run_command(cmd):
    """Prints output from running a command as it occurs.

    Args:
        cmd: Either a string command or array of strings, as one would pass to
            :code:`subprocess.run()`
    """
    logging.debug("Running command '%s'" % str(cmd))

    print(*cmd)

    with subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1, text=True) as p:
        for line in p.stdout:
            print(line, end="")  # process line here


def get_current_commit() -> str:
    """Returns printed output from running :code:`git rev-parse HEAD` command."""
    return get_command_output(["git", "rev-parse", "HEAD"])


def get_pip_freeze() -> str:
    """Returns printed output from running :code:`pip freeze` command."""
    return get_command_output(["pip", "freeze"])


def get_conda_env() -> str:
    """Returns printed output from running :code:`conda env export --from-history` command."""
    output = ""
    execs = ["conda", "mamba", "micromamba"]
    while output == "" and len(execs) > 0:
        cmd_word = execs[0]
        output = get_command_output([cmd_word, "env", "export", "--from-history"], True)
        execs.remove(cmd_word)
    if output == "":
        logging.warning("Unable to run conda or similar command")
        return output
    else:
        # fix random garbage at end of output (?!)
        output = output[:-7]
        return output


def get_conda_list() -> str:
    """Returns printed output from running :code:`conda env export --from-history` command."""
    return get_command_output(["conda", "list"])


def get_os() -> str:
    """Get the current OS name and version."""
    return str(platform.platform())


def get_py_opening_comment(lines):
    """Parse the passed lines of a python script file for a top of file docstring.
    This is used to get parameter and experiment file descriptions, so be sure to
    always comment your code!"""
    content = ""

    def check_for_end(line):
        if '"""' in line or "'''" in line:
            return True
        return False

    def endline_index(line):
        if '"""' in line:
            return line.index('"""')
        else:
            return line.index("'''")

    if len(lines[0]) > 3 and (lines[0].startswith('"""') or lines[0].startswith("'''")):
        if len(lines[0]) > 3 and not check_for_end(lines[0][3:]):
            content += lines[0][3:]

            for line in lines[1:]:
                if not check_for_end(line):
                    content += " " + line
                else:
                    content += " " + line[: endline_index(line)]
                    break
        elif len(lines[0]) > 3 and check_for_end(lines[0][3:]):
            content += lines[0][3 : endline_index(lines[0][3:]) + 3]

    # fix line breaks
    content = content.replace("\n", " ")
    while "  " in content:
        content = content.replace("  ", " ")

    return content


def set_logging_prefix(prefix):
    """Set the prefix content of the logger, which is incorporated in the log formatter. Currently
    this is just used to include pid number when running in parallel."""
    # https://stackoverflow.com/questions/17558552/how-do-i-add-custom-field-to-python-log-format-string
    old_factory = logging.getLogRecordFactory()

    def new_factory(*args, **kwargs):
        record = old_factory(*args, **kwargs)
        record.prefix = prefix
        return record

    logging.setLogRecordFactory(new_factory)


def init_logging(
    log_path=None, level=logging.INFO, log_errors=False, include_process=False
):
    """Sets up logging configuration, including the associated file output.

    Args:
        log_path (str): Folder to store output logs in. If :code:`None`, only log
            to console.
        level: The logging level to output.
        log_errors (bool): Whether to include error messages in the log output or not.
        include_process (bool): Whether to include the PID prefix value in the logger.
            This is mostly only used for when the :code:`--parallel` flag is used, to
            help track which log message is from which process.
    """
    if include_process:
        # NOTE: taking out %(filename)s because it takes up space and makes the beginning of the log entries "jagged"
        log_formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] {PID:%(process)s} - %(prefix)s%(message)s"
        )
    else:
        log_formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] - %(prefix)s%(message)s"
        )
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    root_logger.propagate = False
    root_logger.handlers = []

    logging.addLevelName(logging.DEBUG, "DBUG")

    set_logging_prefix("")

    if log_path is not None:
        file_handler = logging.FileHandler(log_path)
        file_handler.setFormatter(log_formatter)
        root_logger.addHandler(file_handler)

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(log_formatter)
    root_logger.addHandler(console_handler)

    # https://stackoverflow.com/questions/27538879/how-to-disable-loggers-from-other-modules
    # disable all loggers except ours (TODO: may want to config this at some point)
    for name, logger in logging.root.manager.loggerDict.items():
        logger.disabled = True

    # sys.stdout = StreamToLogger(logging.INFO)
    if log_errors:
        sys.stderr = StreamToLogger(logging.ERROR)

    # logging.getLogger("torch").setLevel(logging.WARNING)
    # logging.getLogger("transformers").setLevel(logging.WARNING)
    # logging.getLogger("requests").setLevel(logging.WARNING)
    # logging.getLogger("urllib3").setLevel(logging.WARNING)
    # logging.getLogger("snorkel").setLevel(logging.WARNING)
    # logging.getLogger("snorkel.labeling.model").setLevel(logging.WARNING)
    # logging.getLogger("snorkel.labeling.model.logger").setLevel(logging.WARNING)


# setting store_in_registry to true will store the hash and associated args dictionary
# in registry json file
def args_hash(args, registry_path: str, store_in_registry: bool = False) -> str:
    """Returns a hex string representing the passed arguments, optionally recording
    the arguments and hash in the params registry.

    Args:
        args (ExperimentArgs): The argument set to hash.
        registry_path (str): The location to keep the :code:`params_registry.json`.
        store_in_registry (bool): Whether to update the params registry with the passed
            arguments or not.

    Returns:
        The hash string computed from the arguments.
    """
    if args.hash is not None:
        hash_str = args.hash
    else:
        args_copy = deepcopy(args)

        # get rid of parameters that shouldn't affect hash
        args_dict = asdict(args_copy)
        del args_dict["overwrite"]
        del args_dict["hash"]

        hash_key = str(args_dict)

        hash_str = hashlib.md5(hash_key.encode()).hexdigest()

    def stringify(x):
        # NOTE: at some point it may be worth doing fancier logic here. Objects
        # that don't have __str__ implemented will return a string with a pointer,
        # which will always be different regardless
        return str(x)

    if store_in_registry:
        registry_path = os.path.join(registry_path, "params_registry.json")
        registry = {}

        if os.path.exists(registry_path):
            with open(registry_path, "r") as infile:
                registry = json.load(infile)

        registry[hash_str] = asdict(args)
        with open(registry_path, "w") as outfile:
            json.dump(registry, outfile, indent=4, default=stringify)

    return hash_str


# hash_list should _not_ have the active hash in it
# def add_args_combo_hash(hash_str, active_hash, hash_list, registry_path: str):
def add_args_combo_hash(
    active_record, records_list, registry_path: str, store_in_registry: bool = False
):
    """Returns a hex string representing the the combined argument set hashes from the
    passed records list. This is mainly used for getting a hash for an aggregate stage,
    which may not have a meaningful argument set of its own.

    Args:
        active_record (Record): The currently in-use record (likely owned by the aggregate
            stage.)
        records_list (List[Record]): The list of records to include as part of the resulting
            hash.
        registry_path (str): The location to keep the :code:`params_registry.json`.
        store_in_registry (bool): Whether to update the params registry with the passed
            records or not.

    Returns:
        The hash string computed from the combined record arguments.
    """

    hashes = []
    for agg_record in records_list:
        if agg_record.args is not None:
            hashes.append(agg_record.args.hash)
        else:
            hashes.append("None")
    hashes = sorted(hashes)

    hashes_for_key = deepcopy(hashes)
    active_key = "None"
    if active_record.args is not None:
        active_key = active_record.args.hash
    hashes_for_key.insert(0, active_key)

    hash_key = str(hashes_for_key)
    hash_str = hashlib.md5(hash_key.encode()).hexdigest()

    if store_in_registry:
        registry_path = os.path.join(registry_path, "params_registry.json")
        registry = {}

        if os.path.exists(registry_path):
            with open(registry_path, "r") as infile:
                registry = json.load(infile)

        registry[hash_str] = {"active": active_key, "arg_list": hashes}
        with open(registry_path, "w") as outfile:
            json.dump(registry, outfile, indent=4, default=lambda x: str(x))
    return hash_str


# https://stackoverflow.com/questions/19425736/how-to-redirect-stdout-and-stderr-to-logger-in-python
class StreamToLogger(object):
    def __init__(self, level):
        # self.logger = logger
        self.level = level

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            logging.log(self.level, line.rstrip())



"""This is the 'meta' curifactory CLI, which is a runnable to help set up a
project structure.

This file contains a :code:`__name__ == "__main__"` and can be run directly.
"""

import argparse
import json
import os
import pkg_resources
import shutil

import curifactory as cf
from curifactory import utils


def initialize_project():
    print("Initializing curifactory project...")
    print(
        "Enter paths for configuration file, default values shown in '[]' (leave entry blank to use default)"
    )
    if os.path.exists(utils.CONFIGURATION_FILE):
        print(
            "NOTE: existing curifactory configuration found, default values are now currently existing values."
        )
    config = utils.get_configuration()

    def query_value_and_update(config, query, key):
        value = input(f"{query} ['{config[key]}']: ")
        if value != "":
            config[key] = value

    query_value_and_update(config, "Experiments module name", "experiments_module_name")
    query_value_and_update(config, "Parameters module name", "params_module_name")
    query_value_and_update(config, "Data path", "manager_cache_path")
    config["cache_path"] = os.path.join(config["manager_cache_path"], "cache/")
    config["runs_path"] = os.path.join(config["manager_cache_path"], "runs/")
    query_value_and_update(config, "Logs path", "logs_path")
    query_value_and_update(config, "Notebooks path", "notebooks_path")
    query_value_and_update(config, "Reports path", "reports_path")
    config["report_css_path"] = os.path.join(config["reports_path"], "style.css")

    # write out config
    print("Writing config file...")
    with open("curifactory_config.json", "w") as outfile:
        json.dump(config, outfile, indent=4)

    # create directory structure
    print("Setting up folders...")
    os.makedirs(config["experiments_module_name"], exist_ok=True)
    os.makedirs(config["params_module_name"], exist_ok=True)
    os.makedirs(config["manager_cache_path"], exist_ok=True)
    os.makedirs(config["cache_path"], exist_ok=True)
    os.makedirs(config["runs_path"], exist_ok=True)
    os.makedirs(config["logs_path"], exist_ok=True)
    os.makedirs(config["notebooks_path"], exist_ok=True)
    os.makedirs(os.path.join(config["notebooks_path"], "experiments"), exist_ok=True)
    os.makedirs(config["reports_path"], exist_ok=True)

    # copy in style sheet for reports
    style_path = pkg_resources.resource_filename("curifactory", "data/style.css")
    shutil.copyfile(style_path, config["report_css_path"])

    # handle docker folder and dockerfile
    valid_docker_choice = False
    while not valid_docker_choice:
        docker_yn = input("Include docker folder and default dockerfile? [Y/n] ")
        if docker_yn == "" or docker_yn.lower() == "y":
            valid_docker_choice = True
            print("Setting up docker folder...")
            os.makedirs("docker", exist_ok=True)

            # read in the dockerfile contents and edit appropriately
            dockerfile_path = pkg_resources.resource_filename(
                "curifactory", "data/dockerfile"
            )
            with open(dockerfile_path, "r") as infile:
                contents = infile.read()
                contents.replace("{{CF_VERSION}}", cf.__version__)
            with open("docker/dockerfile", "w") as outfile:
                outfile.write(contents)

            dockerfile_start_path = pkg_resources.resource_filename(
                "curifactory", "data/startup.sh"
            )
            shutil.copyfile(dockerfile_start_path, "docker/startup.sh")

            dockerfile_ignore_path = pkg_resources.resource_filename(
                "curifactory", "data/.dockerignore"
            )
            shutil.copyfile(dockerfile_ignore_path, "docker/.dockerignore")
        elif docker_yn.lower() == "n":
            valid_docker_choice = True
        if not valid_docker_choice:
            print("Invalid entry, please enter 'y' or 'n', or leave blank for default.")

    # handle gitignore
    valid_gitfile_choice = False
    while not valid_gitfile_choice:
        gitfile_yn = input("Append curifactory paths to .gitignore? [Y/n] ")
        if gitfile_yn == "" or gitfile_yn.lower() == "y":
            valid_gitfile_choice = True

            newline_needed = False
            if os.path.exists(".gitignore"):
                with open(".gitignore", "r") as infile:
                    lines = infile.readlines()
                    if len(lines) > 0 and lines[-1] != "\n":
                        newline_needed = True

            print("Appending to .gitignore...")
            with open(".gitignore", "a") as outfile:
                if newline_needed:
                    outfile.write("\n")
                outfile.write(
                    "\n".join(
                        [
                            "# curifactory paths",
                            os.path.join(config["manager_cache_path"], "*"),
                            os.path.join(config["cache_path"], "*"),
                            os.path.join(config["runs_path"], "*"),
                            os.path.join(config["logs_path"], "*"),
                            os.path.join(config["reports_path"], "*"),
                            "!" + config["report_css_path"],
                        ]
                    )
                )
        elif gitfile_yn.lower() == "n":
            valid_gitfile_choice = True
        if not valid_gitfile_choice:
            print("Invalid entry, please enter 'y' or 'n', or leave blank for default.")

    print("Curifactory project initialization complete!")


def main():

    # TODO: could add the reset functionality as requested in this runnable.

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="subparser_name")
    init_parser = subparsers.add_parser("init")  # noqa: F841 -- we might use eventually

    args = parser.parse_args()
    if args.subparser_name == "init":
        initialize_project()


if __name__ == "__main__":
    main()



"""Testing the decorators to help orchestrate caching and
input/output passing through record state between stages."""

import copy
import logging
import os
import pickle
import psutil

import shutil
import time
from typing import List, Union
from functools import wraps

from curifactory.record import Record, ArtifactRepresentation
from curifactory.caching import Lazy, PickleCacher, FileReferenceCacher
from curifactory import utils


# NOTE: resource only exists on unix systems
if os.name != "nt":
    import resource


class InputSignatureError(Exception):
    pass


class OutputSignatureError(Exception):
    pass


class EmptyCachersError(Exception):
    pass


class CachersMismatchError(Exception):
    pass


def _log_stats(
    record,
    pre_cache_time_start,
    pre_cache_time_end,
    pre_mem_usage,
    post_mem_usage,
    exec_time_start=0,
    exec_time_end=0,
    post_cache_time_start=0,
    post_cache_time_end=0,
    pre_max_footprint=0,
    post_max_footprint=0,
):

    pre_cache_time = pre_cache_time_end - pre_cache_time_start
    exec_time = exec_time_end - exec_time_start
    post_cache_time = post_cache_time_end - post_cache_time_start
    cache_time = pre_cache_time + post_cache_time

    mem_change = post_mem_usage - pre_mem_usage

    footprint_change = post_max_footprint - pre_max_footprint

    # TODO: total_time =

    logging.debug(
        "Memory (current usage/max allocated) - %s / %s"
        % (
            utils.human_readable_mem_usage(post_mem_usage),
            utils.human_readable_mem_usage(post_max_footprint),
        )
    )
    logging.debug(
        "Stage memory impact (current/max) - %s / %s"
        % (
            utils.human_readable_mem_usage(mem_change),
            utils.human_readable_mem_usage(footprint_change),
        )
    )
    logging.debug(
        "Timing - execution: %s  caching: %s"
        % (utils.human_readable_time(exec_time), utils.human_readable_time(cache_time))
    )


def stage(  # noqa: C901 -- TODO: will be difficult to simplify...
    inputs: List[str] = None,
    outputs: List[Union[str, Lazy]] = None,
    cachers: List = None,
    suppress_missing_inputs: bool = False,
):
    """Decorator to wrap around a function that represents a single step in an experiment,
    a block with inputs and outputs pertaining to the remainder of that experiment.

    Important:
        Any function wrapped with the stage decorator must take a Record instance as the first
        parameter, followed by the input parameters corresponding to the :code:`inputs` list.

    Note:
        Note that a wrapped function that successfully finds all cached outputs does not
        execute, meaning any reportables that might have been output to the experiment report
        **will not run.** This can be mitigated by putting relevant reportables in a separate
        stage that does not cache anything. This note similarly applies to any other "side effects"
        that might result from stage code execution. Careful design of stages should ensure
        that the effects of stage functions are limited exclusively to the given inputs and
        returned outputs.

    Args:
        inputs (List[str]): A list of variable names that this stage will need from the
            record state. **Note that all inputs listed here must have a corresponding
            input parameter in the function definition line, each with the exact same name
            as in this list.**
        outputs (List[Union[str, Lazy]]): A list of variable names that this stage will return and store
            in the record state. These represent, in order, the tuple of returned values from
            the function being wrapped.
        cachers (List[Cacheable]): An optional list of Cacheable instances ("strategies") to
            apply to each of the return outputs. If specified, for each output, an instance
            of the corresponding cacher is initialized, and the :code:`save()` function is called.
            Before the wrapped function is called, the output path is first checked, and if it
            exists and the current record args are not set to overwrite, the :code:`load()` function
            is called and the wrapped function **does not execute.** Note that caching is all
            or nothing for a single function, you cannot cache only one returned value out of
            several.
        suppress_missing_inputs (bool): If true, any stage inputs that are not found in the record's
            state will be passed in as :code:`None` rather than raising an exception. This can
            be used to make all inputs optional, such as if a stage will be used after different
            sets of previous stages and not all values are necessarily required.

    Example:
        .. code-block:: python

            @stage(inputs=["data", "model"], outputs=["results"], cachers=[JsonCacher])
            def test_model(record: Record, data: pd.DataFrame, model):
                # ...
                return results_dictionary

        Note that from this example, this stage assumes some other stages have output
        :code:`"data"` and :code:`"model"` at some point.
    """

    def decorator(function):
        @wraps(function)
        def wrapper(record: Record, *args, **kwargs):
            # set the logging prefix to the args name
            if record.args is not None:
                utils.set_logging_prefix(f"[{record.args.name}] ")
            else:
                utils.set_logging_prefix("")

            name = function.__name__
            logging.info("-----")
            logging.info("Stage %s", name)
            pre_footprint = 0
            if os.name != "nt":
                pre_footprint = (
                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                )
            pre_mem_usage = psutil.Process().memory_info().rss
            record.manager.current_stage_name = name
            record.stages.append(name)
            record.stage_outputs.append([])
            record.stage_inputs.append([])

            # apply consistent handling
            nonlocal inputs, outputs, cachers
            if inputs is None:
                inputs = []
            if outputs is None:
                outputs = []

            # check for lazy object / cacher mismatch in outputs
            for output in outputs:
                if (
                    type(output) == Lazy
                    and cachers is None
                    and not record.manager.lazy
                    and not record.manager.ignore_lazy
                ):
                    raise OutputSignatureError(
                        "Stage outputs for '%s' contain Lazy objects but do not have cachers specified. Please provide cachers for this stage."
                        % name
                    )

            # replace any non-lazy outputs with lazy or vice-versa depending on manager flags.
            if record.manager.lazy:
                no_cachers = False
                for index, output in enumerate(outputs):
                    if type(output) != Lazy:
                        logging.debug("Forcing lazy cache for '%s'" % output)
                        outputs[index] = Lazy(output)
                        # NOTE: since Lazy caching doesn't work without a cacher, we need to ensure
                        # one if none exists. Pickle is pretty broad, but obviously there are some things
                        # that don't work, so we need to warn about this
                        if cachers is None:
                            no_cachers = True
                            logging.warning(
                                "Stage %s does not have cachers specified, a --lazy run will force caching by applying PickleCachers to anything with none specified, but this can potentially cause errors."
                                % name
                            )
                            cachers = []
                        if no_cachers:
                            cachers.append(PickleCacher)
            elif record.manager.ignore_lazy:
                for index, output in enumerate(outputs):
                    if type(output) == Lazy:
                        logging.debug("Disabling lazy cache for '%s'" % output)
                        outputs[index] = output.name

            # check for mismatched amounts of cachers
            if cachers is not None and len(cachers) != len(outputs):
                raise CachersMismatchError(
                    f"Stage '{name}' - the number of cachers does not match the number of outputs to cache."
                )

            # find any required inputs for the function in the record
            # NOTE: this is before checking cache so that we can keep a record of what inputs we're expecting
            # (for mapping purposes)
            # note that we turn off lazy resolution because we haven't yet
            # determined if the stage even needs to run yet or not. We handle
            # resolution manually farther down.
            record.state.resolve = False
            function_inputs = {}
            for function_input in inputs:
                # NOTE: this means the state was probably manually changed
                if function_input not in record.state_artifact_reps:
                    record.stage_inputs[-1].append(-1)
                else:
                    record.stage_inputs[-1].append(
                        record.state_artifact_reps[function_input]
                    )
                # TODO: add None if function_input is not in state. (There are a few cases
                # where this shouldn't be a problem, e.g. if everything is cached but a new
                # input was added - this shouldn't inherently cause a problem unless the
                # output would be wrong.
                if function_input not in record.state:
                    if suppress_missing_inputs:
                        logging.warning(
                            "Suppressed missing inputs, will expect function signature default value for '%s' or a direct argument pass on the stage function call..."
                            % function_input
                        )
                        # function_inputs[function_input] = None
                        # NOTE: we don't actually need to pass in None, we can expect the user to implement whatever defaults they want for optional parameters. If they don't specify a default, this will fail as normal.
                    elif function_input not in kwargs:
                        raise KeyError(
                            "Stage '%s' input '%s' not found in record state and not passed to function call. Set 'suppress_missing_inputs=True' on the stage and give a default value in the function signature if this should run anyway."
                            % (name, function_input)
                        )
                else:
                    # # check for a lazy object and resolve it if so
                    # if type(record.state[function_input]) == Lazy:
                    #     # TODO: no no no don't do this here, we haven't checked for cached outputs yet.
                    #     logging.debug("Resolving lazy load object '%s'" % function_input)
                    #     function_inputs[function_input] = record.state[function_input].load()
                    # else:
                    #     # otherwise we just directly pull it from the record state.
                    function_inputs[function_input] = record.state[function_input]
            function_inputs.update(kwargs)
            record.state.resolve = True

            # note to future self and anyone else who's IDE says this is repeated code (with aggregate below)
            # no, you cannot abstract this into _check_cached_outputs - if you try to reassign to cachers from
            # another function, because of the deep voodoo black magic sorcery that is decorators with arguments,
            # it considers it different code.
            if cachers is not None:
                # instantiate cachers if not already
                for i in range(len(cachers)):
                    cacher = cachers[i]
                    if type(cacher) == type:
                        cachers[i] = cacher()
                    # set the active record on the cacher
                    cachers[i].record = record

            # check for cached outputs and lazy load inputs if needed
            pre_cache_time_start = time.perf_counter()  # time to load from cache
            record.manager.lock()
            cache_valid = _check_cached_outputs(name, record, outputs, cachers)
            if cache_valid:
                # get previous reportables if available
                _check_cached_reportables(name, record)

            # check each input for Lazy objects and load them if we know we have to execute this stage
            if not cache_valid:
                for function_input in function_inputs:
                    if type(function_inputs[function_input]) == Lazy:
                        logging.debug(
                            "Resolving lazy load object '%s'" % function_input
                        )
                        function_inputs[function_input] = function_inputs[
                            function_input
                        ].load()

            record.manager.unlock()
            pre_cache_time_end = time.perf_counter()
            if cache_valid:
                post_mem_usage = psutil.Process().memory_info().rss
                post_footprint = 0
                if os.name != "nt":
                    post_footprint = (
                        resource.getrusage(resource.RUSAGE_THREAD).ru_maxrss * 1024
                    )
                _log_stats(
                    record,
                    pre_cache_time_start,
                    pre_cache_time_end,
                    pre_mem_usage,
                    post_mem_usage,
                    0,
                    0,
                    0,
                    0,
                    pre_footprint,
                    post_footprint,
                )
                utils.set_logging_prefix("")
                return record

            # run the function
            logging.info("Stage %s executing...", name)
            exec_time_start = time.perf_counter()
            try:
                function_outputs = function(record, *args, **function_inputs)
            except TypeError as e:
                raise InputSignatureError(
                    "Signature for '%s' does not match stage input list. Signature should include %s, or there may be missing default values for a stage called with suppress_missing_inputs. Sub error: %s"
                    % (name, str(inputs), str(e))
                )
            exec_time_end = time.perf_counter()

            # handle storing outputs in record
            post_cache_time_start = time.perf_counter()
            record.manager.lock()
            _store_outputs(name, record, outputs, cachers, function_outputs)
            _store_reportables(name, record)
            record.manager.unlock()
            post_cache_time_end = time.perf_counter()

            # free up any memory from cached things
            cleaned_function_outputs = []
            lazy_found = False

            # iterate the outputs rather than function_outputs
            for index, output_name in enumerate(outputs):
                if type(output_name) == Lazy:
                    lazy_found = True
                    logging.debug("Lazy object '%s' will be cleaned." % output_name)
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = outputs[
                            index
                        ]  # so that this doesn't syntactically read output_name, because it's Lazy not a str
                    else:
                        cleaned_function_outputs.append(outputs[index])
                else:
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = function_outputs
                    else:
                        cleaned_function_outputs.append(function_outputs[index])
            if len(outputs) > 1:
                cleaned_function_outputs = tuple(cleaned_function_outputs)

            # free up any lazy cache objects
            if lazy_found:
                logging.debug("Freeing memory from lazy objects...")
                pre_del_mem_usage = psutil.Process().memory_info().rss
                del function_outputs
                post_del_mem_usage = psutil.Process().memory_info().rss
                del_mem_diff = pre_del_mem_usage - post_del_mem_usage
                logging.debug("Freed %s" % utils.human_readable_mem_usage(del_mem_diff))

            logging.info("Stage %s complete", name)

            # check memory usage
            post_footprint = 0
            if os.name != "nt":
                post_footprint = (
                    resource.getrusage(resource.RUSAGE_THREAD).ru_maxrss * 1024
                )
            post_mem_usage = psutil.Process().memory_info().rss

            _log_stats(
                record,
                pre_cache_time_start,
                pre_cache_time_end,
                pre_mem_usage,
                post_mem_usage,
                exec_time_start,
                exec_time_end,
                post_cache_time_start,
                post_cache_time_end,
                pre_footprint,
                post_footprint,
            )

            record.output = cleaned_function_outputs
            utils.set_logging_prefix("")
            return record

        return wrapper

    return decorator


def aggregate(  # noqa: C901 -- TODO: will be difficult to simplify...
    outputs: List[str] = None, cachers: List = None
):
    """Decorator to wrap around a function that represents some step that must operate across
    multiple different argsets or "experiment lines" within an experiment. This is normally
    used to run final analyses and comparisons of results across all passed argument sets.

    Important:
        Any function wrapped with the aggregate decorator must take a Record instance as the first
        parameter and a list of Record instances as the second. The former is the record that applies
        to this function, and the latter is the set of other records from elsewhere in the experiment
        that this function needs to aggregate across.

    Args:
        outputs (List[str]): A list of variable names that this stage will return and store
            in the record state. These represent, in order, the tuple of returned values from
            the function being wrapped.
        cachers (List[Cacheable]): An optional list of Cacheable instances ("strategies") to
            apply to each of the return outputs. If specified, for each output, an instance
            of the corresponding cacher is initialized, and the :code:`save()` function is called.
            Before the wrapped function is called, the output path is first checked, and if it
            exists and the current record args are not set to overwrite, the :code:`load()` function
            is called and the wrapped function **does not execute.** Note that caching is all
            or nothing for a single function, you cannot cache only one returned value out of
            several.


    Example:
        .. code-block:: python

            @aggregate(["final_results"], [JsonCacher])
            def compile_results(record: Record, records: List[Record]):
                results = {}
                for prev_record in records:
                    results[prev_record.args.name] = prev_record.state["results"]
                return results
    """

    def decorator(function):
        @wraps(function)
        def wrapper(record: Record, records: List[Record] = None, **kwargs):
            # set the logging prefix to the args name
            if record.args is not None:
                utils.set_logging_prefix(f"[{record.args.name}] ")
            else:
                utils.set_logging_prefix("")

            name = function.__name__
            logging.info("-----")
            logging.info("Stage (aggregate) %s", name)
            pre_footprint = 0
            if os.name != "nt":
                pre_footprint = (
                    resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
                )
            pre_mem_usage = psutil.Process().memory_info().rss
            record.manager.current_stage_name = name
            record.stages.append(name)
            record.stage_outputs.append([])
            record.stage_inputs.append([])

            # apply consistent handling
            nonlocal outputs, cachers
            if outputs is None:
                outputs = []

            # check for lazy object / cacher mismatch in outputs
            for output in outputs:
                if (
                    type(output) == Lazy
                    and cachers is None
                    and not record.manager.lazy
                    and not record.manager.ignore_lazy
                ):
                    raise OutputSignatureError(
                        "Stage outputs for '%s' contain Lazy objects but do not have cachers specified. Please provide cachers for this stage."
                        % name
                    )

            # replace any non-lazy outputs with lazy or vice-versa depending on manager flags.
            if record.manager.lazy:
                no_cachers = False
                for index, output in enumerate(outputs):
                    if type(output) != Lazy:
                        logging.debug("Forcing lazy cache for '%s'" % output)
                        outputs[index] = Lazy(output)
                        # NOTE: since Lazy caching doesn't work without a cacher, we need to ensure
                        # one if none exists. Pickle is pretty broad, but obviously there are some things
                        # that don't work, so we need to warn about this
                        if cachers is None:
                            no_cachers = True
                            logging.warning(
                                "Aggregate stage %s does not have cachers specified, a --lazy run will force caching by applying PickleCachers to anything with none specified, but this can potentially cause errors."
                                % name
                            )
                            cachers = []
                        if no_cachers:
                            cachers.append(PickleCacher)
            elif record.manager.ignore_lazy:
                for index, output in enumerate(outputs):
                    if type(output) == Lazy:
                        logging.debug("Disabling lazy cache for '%s'" % output)
                        outputs[index] = output.name

            if records is None:
                # if no explicit request for a specific set of records given,
                # use all previous records
                records = record.manager.records
            record.input_records = records

            if cachers is not None and len(cachers) != len(outputs):
                raise CachersMismatchError(
                    f"Stage '{name}' - the number of cachers does not match the number of outputs to cache"
                )

            # see note in stage
            if cachers is not None:
                # instantiate cachers if not already
                for i in range(len(cachers)):
                    cacher = cachers[i]
                    if type(cacher) == type:
                        cachers[i] = cacher()
                    # set the active record on the cacher
                    cachers[i].record = record

            pre_cache_time_start = time.perf_counter()  # time to load from cache
            record.manager.lock()
            cache_valid = _check_cached_outputs(name, record, outputs, cachers, records)
            if cache_valid:
                # get previous reportables if available
                _check_cached_reportables(name, record, records)
            record.manager.unlock()
            pre_cache_time_end = time.perf_counter()

            if cache_valid:
                post_mem_usage = psutil.Process().memory_info().rss
                post_footprint = 0
                if os.name != "nt":
                    post_footprint = (
                        resource.getrusage(resource.RUSAGE_THREAD).ru_maxrss * 1024
                    )
                _log_stats(
                    record,
                    pre_cache_time_start,
                    pre_cache_time_end,
                    pre_mem_usage,
                    post_mem_usage,
                    0,
                    0,
                    0,
                    0,
                    pre_footprint,
                    post_footprint,
                )
                utils.set_logging_prefix("")
                return record

            # run the function
            logging.info("Stage (aggregate) %s executing...", name)
            exec_time_start = time.perf_counter()
            # NOTE: passing additional args to an aggregate stage is sort of undefined functionality,
            # this probably should be discouraged.
            function_outputs = function(record, records, **kwargs)
            exec_time_end = time.perf_counter()

            # handle storing outputs in record
            post_cache_time_start = time.perf_counter()
            record.manager.lock()
            _store_outputs(name, record, outputs, cachers, function_outputs, records)
            _store_reportables(name, record, records)
            record.manager.unlock()
            post_cache_time_end = time.perf_counter()

            # free up any memory from cached things
            cleaned_function_outputs = []
            lazy_found = False

            # iterate the outputs rather than function_outputs
            for index, output_name in enumerate(outputs):
                if type(output_name) == Lazy:
                    lazy_found = True
                    logging.debug("Lazy object '%s' will be cleaned." % output_name)
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = outputs[
                            index
                        ]  # so that this doesn't syntactically read output_name, because it's Lazy not a str
                    else:
                        cleaned_function_outputs.append(outputs[index])
                else:
                    if index == 0 and len(outputs) == 1:
                        cleaned_function_outputs = function_outputs
                    else:
                        cleaned_function_outputs.append(function_outputs[index])
            if len(outputs) > 1:
                cleaned_function_outputs = tuple(cleaned_function_outputs)

            # free up any lazy cache objects
            if lazy_found:
                logging.debug("Freeing memory from lazy objects...")
                pre_del_mem_usage = psutil.Process().memory_info().rss
                del function_outputs
                post_del_mem_usage = psutil.Process().memory_info().rss
                del_mem_diff = pre_del_mem_usage - post_del_mem_usage
                logging.debug("Freed %s" % utils.human_readable_mem_usage(del_mem_diff))

            logging.info("Stage (aggregate) %s complete", name)

            # check memory usage
            post_footprint = 0
            if os.name != "nt":
                post_footprint = (
                    resource.getrusage(resource.RUSAGE_THREAD).ru_maxrss * 1024
                )
            post_mem_usage = psutil.Process().memory_info().rss

            _log_stats(
                record,
                pre_cache_time_start,
                pre_cache_time_end,
                pre_mem_usage,
                post_mem_usage,
                exec_time_start,
                exec_time_end,
                post_cache_time_start,
                post_cache_time_end,
                pre_footprint,
                post_footprint,
            )

            record.output = cleaned_function_outputs
            utils.set_logging_prefix("")
            return record

        return wrapper

    return decorator


def _check_cached_outputs(stage_name, record, outputs, cachers, records=None):
    if cachers == []:
        raise EmptyCachersError(
            "Do not use '[]' for cachers. This will always short-circuit because there is nothing that isn't cached."
        )

    cache_valid = False
    if cachers is not None:
        # set the path for every instantiated cacher
        paths = []
        for index, arg in enumerate(outputs):
            if cachers[index].path_override is None:
                # the str(arg) will handle Lazy objects
                path = record.manager.get_path(
                    str(arg), record, aggregate_records=records
                )
            elif str.endswith(cachers[index].extension, cachers[index].path_override):
                # if the path override includes the extension they provided a full file name
                # NOTE: this is useful if there's a static file that won't change across diff
                # runs or paramsets, like an input dataset
                path = cachers[index].path
            else:
                path = record.manager.get_path(
                    arg,
                    record,
                    base_path=cachers[index].path,
                    aggregate_records=records,
                )
            path = cachers[index].set_path(path)
            paths.append(path)

        # NOTE: we put this here because the cachers still need to be set up
        if stage_name in record.manager.overwrite_stages or record.manager.overwrite:
            return False

        # check the cache (and load into record if found)
        cache_valid = True
        function_outputs = []
        for i in range(len(paths)):
            if cachers[i].check():

                # handle lazy objects by setting the cacher but not actually loading yet.
                if type(outputs[i]) == Lazy:
                    outputs[i].cacher = cachers[i]
                    # we set the output to just be the Lazy instance for now
                    output = outputs[i]
                else:
                    output = cachers[i].load()
                function_outputs.append(output)
                record.state[str(outputs[i])] = output

                artifact = _add_output_artifact(record, output, outputs, i)
                artifact.file = cachers[i].path

                # copy it over to output run folder if necessary
                if record.manager.store_entire_run:
                    # if we don't handle lazy separately it will literally store the lazy object.
                    # Instead, just use the OS to copy the file over. (This avoids us having to
                    # eat the memory costs of reloading and resaving.)
                    previous_path = cachers[i].path
                    cachers[i].set_path(
                        record.manager.get_path(
                            outputs[i], record, output=True, aggregate_records=records
                        )
                    )
                    if type(outputs[i]) == Lazy:
                        shutil.copyfile(previous_path, cachers[i].path)
                    else:
                        cachers[i].save(output)
            else:
                # we found something that wasn't cached, recompute everything
                cache_valid = False
                break

        # if we have the cached objects, return them right away
        if cache_valid:
            if len(paths) == 1:
                record.output = function_outputs[0]
            else:
                record.output = tuple(function_outputs)
            # return record

    return cache_valid


def _add_output_artifact(record, object, outputs, index):
    """manage representation recording"""
    artifact = ArtifactRepresentation(record, outputs[index], object)
    new_index = len(record.manager.artifacts)
    record.manager.artifacts.append(artifact)
    # if new_index not in record.stage_outputs[-1]:
    record.stage_outputs[-1].append(new_index)
    record.state_artifact_reps[str(outputs[index])] = new_index
    return artifact


def _check_cached_reportables(stage_name, record, aggregate_records=None):
    reportables_list_cacher = FileReferenceCacher()
    reportables_list_cacher.record = record
    reportables_list_cacher.set_path(
        record.manager.get_path(
            "reportables_file_list", record, aggregate_records=aggregate_records
        )
    )
    if reportables_list_cacher.check():
        paths = reportables_list_cacher.load()
        for path in paths:
            with open(path, "rb") as infile:
                logging.debug("Reusing cached reportable '%s'" % path)
                reportable = pickle.load(infile)
                record.report(reportable)
        return True
    # the return of this function is simply used to determine if we need to store again?
    return False


def _store_reportables(stage_name, record, aggregate_records=None):
    # get all reportables from the manager for this record and stage name
    reportables = []
    for reportable in record.manager.reportables:
        if reportable.record == record and reportable.stage == stage_name:
            reportables.append(reportable)

    if len(reportables) == 0:
        return

    # pickle each one and store it. (we'll have to handle store-full the same way as outputs below I think)
    # TODO: (02/10/2022) like record get_dir and get_path normally, _this does not transfer into a store-full
    # run.
    paths = []
    reportables_path = record.get_dir("reportables")
    for reportable in reportables:
        # make a copy of the reportable without the record, because that seems to break the mp.lock
        # when in parallel mode.
        # NOTE: do NOT use a deepcopy below, runs into same issue.
        reportable_copy = copy.copy(reportable)
        reportable_copy.record = None
        reportable_path = os.path.join(reportables_path, f"{reportable.name}.pkl")
        paths.append(reportable_path)
        logging.debug("Caching reportable '%s'" % reportable_path)
        with open(reportable_path, "wb") as outfile:
            pickle.dump(reportable_copy, outfile)

    # write a cache file out containing the reportables path names. This is a...file reference cacher...can we re-use the logic?
    reportables_list_cacher = FileReferenceCacher()
    reportables_list_cacher.record = record
    reportables_list_cacher.set_path(
        record.manager.get_path(
            "reportables_file_list", record, aggregate_records=aggregate_records
        )
    )
    reportables_list_cacher.save(paths)
    # NOTE: unnecessary because the reportables don't get copied over anyway, see todo note above.
    # (we should get this for free without needing this code when we add extra path tracking to the manager.)
    # if record.manager.store_entire_run:
    #     reportables_list_cacher.set_path(
    #         record.manager.get_path(
    #             "reportables_file_list",
    #             record,
    #             output=True,
    #             aggregate_records=aggregate_records,
    #         )
    #     )
    # reportables_list_cacher.save(reportables_path)


def _store_outputs(
    function_name, record, outputs, cachers, function_outputs, records=None
):
    if len(outputs) == 0:
        return

    if cachers is not None:
        logging.info("Stage %s caching outputs..." % function_name)

    if type(function_outputs) != tuple:
        function_outputs = (function_outputs,)

    if len(outputs) != len(function_outputs):
        raise OutputSignatureError(
            "Returned values from '%s' do not match expected stage outputs. The function should return values for %s"
            % (function_name, str(outputs))
        )

    # store each argument in the record and cache if requested
    for index, output in enumerate(function_outputs):
        if type(outputs[index]) == Lazy:
            record.state[str(outputs[index])] = outputs[index]
            # TODO: (01/13/2022) if cachers is none, throw an error
        else:
            record.state[str(outputs[index])] = output

        # manage representation recording
        artifact = _add_output_artifact(record, output, outputs, index)
        if (
            cachers is not None
            and not record.manager.dry
            and not record.manager.dry_cache
        ):
            logging.debug(
                "Caching %s to '%s'..." % (outputs[index], cachers[index].path)
            )
            cachers[index].save(output)
            artifact.file = cachers[index].path

            # check if we store an additional run output copy
            if record.manager.store_entire_run:
                cachers[index].set_path(
                    record.manager.get_path(
                        outputs[index], record, output=True, aggregate_records=records
                    )
                )
                logging.debug(
                    "Caching %s to '%s'..." % (outputs[index], cachers[index].path)
                )
                cachers[index].save(output)

        # if specified as lazy, be sure to populate the cacher
        if type(outputs[index]) == Lazy:
            outputs[index].cacher = cachers[index]



import curifactory as cf
import json
import os
import pytest

from stages.cache_stages import filerefcacher_stage, filerefcacher_stage_multifile


# TODO: necessary? configured_test_manager already does this
@pytest.fixture()
def clear_stage_run(configured_test_manager):
    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")
    try:
        os.remove(ran_path)
    except FileNotFoundError:
        pass
    yield
    try:
        os.remove(ran_path)
    except FileNotFoundError:
        pass


def test_filerefcacher_stores_multiple_paths(configured_test_manager, clear_stage_run):
    """FileReferenceCacher should correctly store a list of files in the saved json."""
    r = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    filerefcacher_stage_multifile(r)

    argshash = r.args.hash
    expected_list = [
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{argshash}_filerefcacher_stage_multifile_my_files/thing{i}",
        )
        for i in range(5)
    ]

    with open(
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{argshash}_filerefcacher_stage_multifile_output_paths.json",
        ),
        "r",
    ) as infile:
        filelist = json.load(infile)
        assert filelist == expected_list

        for filename in filelist:
            assert os.path.exists(filename)


def test_filerefcacher_stores_single_path(configured_test_manager, clear_stage_run):
    r = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    filerefcacher_stage(r)

    argshash = r.args.hash
    expected_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{argshash}_filerefcacher_stage_my_file",
    )

    with open(
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{argshash}_filerefcacher_stage_output_path.json",
        ),
        "r",
    ) as infile:
        filelist = json.load(infile)
        assert filelist == expected_path


def test_filerefcacher_shortcircuits(configured_test_manager, clear_stage_run):
    """FileReferenceCacher should short-circuit if all files in the filelist already exist."""
    r0 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    filerefcacher_stage_multifile(r0)

    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")
    assert os.path.exists(ran_path)
    os.remove(ran_path)

    r1 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    filerefcacher_stage_multifile(r1)
    assert not os.path.exists(ran_path)


def test_filerefcacher_runs_when_file_missing(configured_test_manager, clear_stage_run):
    """FileReferenceCacher should _not_ short-circuit if any of the files in the filelist are missing."""
    r0 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    filerefcacher_stage_multifile(r0)

    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")
    assert os.path.exists(ran_path)
    os.remove(ran_path)
    os.remove(
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{r0.args.hash}_filerefcacher_stage_multifile_my_files/thing1",
        )
    )

    r1 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    filerefcacher_stage_multifile(r1)
    assert os.path.exists(ran_path)



"""Classes for handling reporting - adding customizable pieces of information
to an HTML experiement run report.

This is handled through a base :code:`Reportable` class, and each reporter class
extends it.
"""

from dataclasses import asdict
import datetime
import json
import logging
import os
import shutil
from typing import List

import matplotlib.pyplot as plt
import numpy as np

from graphviz import Digraph
from graphviz.backend import ExecutableNotFound
import pandas as pd

from curifactory import utils


COLORS = [
    "darkseagreen2",  # #b4eeb4
    "thistle",
    "peachpuff",
    "paleturquoise2",
    "salmon",
    "silver",
    "hotpink",
    "deepskyblue",
    "yellowgreen",
]

COLOR_VALS = {  # because apparently some of graphviz color names aren't websafe
    "darkseagreen2": "#b4eeb4",
    "thistle": "#d8bfd8",
    "peachpuff": "#ffdab9",
    "paleturquoise2": "#aeeeee",
    "salmon": "#fa8072",
    "silver": "#c0c0c0",
    "hotpink": "#ff69b4",
    "deepskyblue": "#00bfff",
    "yellowgreen": "#9acd32",
}


class Reportable:
    """The base reporter class, any custom reporter should extend this.

    Args:
        name (str): A (optional) reference name to give this piece of reported info, it is used
            as the title string suffix. If :code:`None` is supplied, it will be suffixed
            with the number of the reportable.
        group (str): An optional string to use for grouping multiple related reportables
            together in the report. By default, all reportables are ordered by record. This
            will create a separate entry on the TOC and put them next to each other in the
            report.

    Note:
        When subclassing a reportable, :code:`html()` must be overriden, and :code:`render()`
        optionally may be depending on the nature of the reportable. If a reportable relies on
        some form of external file, such as an image or figure, implement :code:`render()` to save
        it (using this class's :code:`path` variable as the directory), and then reference it
        in the output from :code:`html()`. The internal reporting mechanisms handle calling
        both of these functions as needed.

        A simplified example of the :code:`FigureReporter` is shown here:

        .. code-block:: python

            class FigureReporter(Reportable):
                def __init__(self, fig, name=None, group=None):
                    self.fig = fig
                    super().__init__(name=name, group=group)

                def render(self):
                    self.fig.savefig(os.path.join(self.path, f"{self.name}.png"))

                def html(self):
                    return f"<img src='{self.path}/{self.name}.png'>"
    """

    def __init__(self, name=None, group=None):
        self.rendered: bool = False
        """A flag indicating whether this reportable's :code:`render()` has been called yet or not."""
        self.path: str = ""
        """Set internally by reporting functions, this variable holds a valid path where a
        reportable can save files (e.g. images) as needed. This is available to access both
        in :code:`render()` and :code:`html()`"""
        self.name: str = name
        """The suffix to title the reportable with."""
        self.group: str = group
        """If specified, reports group all reportables with the same :code:`group` value together."""
        self.record = None
        """Record: The record this reportable came from, automatically populated via :code:`record.report()`"""
        self.stage: str = ""
        """The name of the stage this reportable comes from, used as part of the title."""

    def html(self):
        """When a report is created, the :code:`html()` function for every reportable is
        called and appended to the report. This function should either return a single
        string of html, or can return a list of lines of html.

        Note:
            Any subclass is **required** to implement this.
        """
        pass

    def render(self):
        """Any file outputs or calculations that should only run once go here."""
        pass


class HTMLReporter(Reportable):
    """Adds the raw string of HTML passed to it to the report.

    Args:
        html_string (str): The raw string of HTML to include.

    Example:
        .. code-block:: python

            @stage(...)
            def report_hello(record: Record ,...):
                record.report(HTMLReporter("<h1>Hello world!</h1>"))

    """

    def __init__(self, html_string, name=None, group=None):
        self.html_string = html_string
        """The raw string of HTML to include."""
        super().__init__(name=name, group=group)

    def html(self):
        return self.html_string


class DFReporter(Reportable):
    """Adds an HTML table to the report for the given pandas dataframe.

    Args:
        df (pd.DataFrame): The pandas dataframe to include in the report.
        float_prec (int): the floating point precision to round all values to.

    Note:
        If you need to use any of the pandas stylers, apply them and pass :code:`df.render()` to
        an :code:`HTMLReporter` instead.
    """

    def __init__(self, df: pd.DataFrame, name=None, group=None, float_prec=4):
        self.df = df
        self.float_prec = float_prec
        super().__init__(name=name, group=group)

    # def html(self):
    #     with pd.option_context("display.precision", self.float_prec, 'display.max_columns', 500, 'display.max_columns')
    #     pd.set_option("display.precision", self.float_prec)

    # def render(self):
    #     # doing this so easily browsable in excel
    #     self.df.to_csv(f"{self.path}{self.name}.csv")

    def html(self):
        output = ["<table border='1' cellspacing='0'><tr><th></th>"]

        # column row
        for col in self.df.columns:
            output.append(f"<th>{col}</th>")

        output.append("</tr>")

        for index, row in self.df.iterrows():
            output.append("<tr>")
            output.append(f"<th>{index}</th>")

            for item in row:
                if (
                    type(item) == float
                    or type(item) == np.float64
                    or type(item) == np.float32
                ):
                    output.append(
                        "<td align='right'><pre>{0:.{1}f}</pre></td>".format(
                            item, self.float_prec
                        )
                    )
                else:
                    output.append(f"<td align='right'><pre>{item}</pre></td>")
            output.append("</tr>")

        output.append("</table>")
        return output


class JsonReporter(Reportable):
    """Adds an indented JSON dump in a :code:`<pre>` tag for a passed dictionary.

    Args:
        dictionary (Dict): The python dictionary to write to a JSON string.
    """

    def __init__(self, dictionary, name=None, group=None):
        self.data = dictionary
        super().__init__(name=name, group=group)

    def html(self):
        return [
            "<pre>",
            json.dumps(self.data, indent=4, default=lambda x: str(x)),
            "</pre>",
        ]


class FigureReporter(Reportable):
    """Adds a passed matplotlib figure to the report.

    Args:
        fig: A matplotlib figure to render.
        kwargs: All keywords args are passed to the figures :code:`savefig()` call in render.
    """

    def __init__(self, fig, name=None, group=None, **kwargs):
        self.kwargs = kwargs
        self.fig = fig
        super().__init__(name=name, group=group)

    def render(self):
        if "format" not in self.kwargs:
            self.kwargs["format"] = "png"
        self.fig.savefig(
            f"{self.path}/{self.name}.{self.kwargs['format']}", **self.kwargs
        )

    def html(self):
        return f"<img src='{self.path}/{self.name}.{self.kwargs['format']}'>"


class LinePlotReporter(Reportable):
    """Takes set(s) of data, creates matplotlib line plots for it, and adds to the report.

    There are several ways to use this reportable, x (optional) and y are both either a single
    list/numpy array of data, or are both dictionaries of lists/numpy arrays of data to plot.

    Possible combinations:

    * Passing a single list/numpy array of only y data
    * Passing a single list/numpy array of y and x data
    * Passing a dictionary (where keys appear in the legend) of lists/numpy arrays of y data, each to be plotted as its own line.
    * Passing a dictionary (where keys appear in the legend) of lists/numpy arrays of both y and x data, each to be plotted as its own line.

    Args:
        y: a single list/numpy array or dictionary of lists/numpy arrays of y data.
        x: (optional), a single list/numpy array or dictionary of lists/numpy arrays of x data.
            If specified, this must match y.
        plot_kwargs (Dict): The **kwargs to pass to the matplotlib :code:`plt.plot()` call.
        savefig_kwargs (Dict): The **kwargs to pass to the :code:`fig.savefig()` call on render.

    Example:
        .. code-block:: python

            @stage(...)
            def report_results(record: Record, ...):

                train_loss = [0.8, 0.7, 0.6]
                test_loss = [0.5, 0.3, 0.4]
                acc = [.93, 0.95, 0.96]

                record.report(LinePlotReporter(
                    y={
                        "Training loss": train_loss,
                        "Testing loss": test_loss,
                        "Accuracy": acc
                    },
                    name="training_history"))

        The above block will result in the following graph added to the report:

        .. figure:: line_plot_reporter_example.png
            :align: center
    """

    def __init__(
        self, y, x=None, name=None, group=None, plot_kwargs={}, savefig_kwargs={}
    ):
        self.savefig_kwargs = savefig_kwargs
        self.plot_kwargs = plot_kwargs
        self.y = y
        self.x = x
        super().__init__(name=name, group=group)

    def render(self):
        plt.figure(facecolor="white")

        if self.x is None:
            if isinstance(self.y, dict):
                for key in self.y:
                    plt.plot(self.y[key], label=key, **self.plot_kwargs)
                plt.legend()
            else:
                plt.plot(self.y, **self.plot_kwargs)
        else:
            if isinstance(self.y, dict):
                for key in self.y:
                    plt.plot(self.x[key], self.y[key], label=key, **self.plot_kwargs)
            else:
                plt.plot(self.x, self.y, **self.plot_kwargs)
        plt.grid(True)

        fig = plt.gcf()
        if "format" not in self.savefig_kwargs:
            self.savefig_kwargs["format"] = "png"
        fig.savefig(
            f"{self.path}/{self.name}.{self.savefig_kwargs['format']}",
            **self.savefig_kwargs,
        )
        plt.close()

    def html(self):
        return f"<img src='{self.path}/{self.name}.{self.savefig_kwargs['format']}'>"


def render_report_head(manager) -> List[str]:
    """Generates the report head tag."""
    return [
        f"<head><title>{manager.experiment_name}/{manager.experiment_run_number}</title>",
        "<link rel='stylesheet' href='style.css'></head>",
    ]


def render_report_info_block(manager) -> List[str]:
    """Generate the header and block of metadata at the top of the report."""

    html_lines = []

    status_color = ""
    if manager.status == "incomplete":
        status_color = "orange"
    elif manager.status == "complete":
        status_color = "green"
    elif manager.status == "error":
        status_color = "red"
    elif manager.status == "LIVE":
        status_color = "cyan"
    status_line = (
        f"<b><span style='color: {status_color}'>{manager.status.upper()}</span></b>"
    )
    if manager.status == "error":
        status_line += " - " + manager.error
    if manager.interactive:
        status_line += " - (interactive session)"

    html_lines.append(
        f"<h1 id='title'>Report: {manager.experiment_name} - {manager.experiment_run_number}</h1>"
    )
    html_lines.extend(
        [
            "<div id='experiment-info-block'>",
            f"<p>Experiment name: <b>{manager.experiment_name}</b> </br>",
            f"Experiment run number: <b>{manager.experiment_run_number}</b></br>",
            f"Run timestamp: <b>{manager.run_timestamp.strftime('%m/%d/%Y %H:%M:%S')}</b></br>",
            f"Reference: <b>{manager.get_reference_name()}</b></br>",
            f"Hostname: <b>{manager.hostname}</b></br>",
            f"Run status: {status_line}</br>",
            f"Git commit: {manager.git_commit_hash}</br>",
            f"Params files: {str(manager.experiment_args_file_list)}</br></p>",
            "<ul>",
        ]
    )
    # output the list of parameters used and assoc hashes
    handled_hashes = []
    for key in manager.experiment_args:
        html_lines.append(f"<li>{key} <ul>")
        for name, args_hash in manager.experiment_args[key]:
            html_lines.append(f"<li>{name} - {args_hash}</li>")
            handled_hashes.append(args_hash)
        html_lines.append("</ul></li>")
    handled_non_params_header = (
        False  # did we print out a header for non-file based params yet?
    )
    for argset in manager.get_all_argsets():
        if argset.hash not in handled_hashes:
            if not handled_non_params_header:
                html_lines.append("<li>Non-file (live) argsets <ul>")
                handled_non_params_header = True
            html_lines.append(f"<li>{argset.name} - {argset.hash}</li>")
    if handled_non_params_header:
        html_lines.append("</ul></li>")

    html_lines.append("</ul></div>")

    if manager.status == "LIVE":
        # TODO: (37/26/2022) allow suppression if notebook=true passed so can render in notebook without error?
        html_lines.append(
            "<p><span style='color: orange;'><b>WARNING - </b></span>This report was not generated from an experiment script. If generated from an interactive environment, curifactory has no method to reproduce.</p>"
        )
        html_lines.append(
            f"<p id='run-string'>Run string: <pre>{manager.run_line}</pre></p>"
        )

        if manager.store_entire_run:
            store_entire_run_path = os.path.join(
                manager.runs_path, manager.get_reference_name()
            )
            html_lines.append(
                f"<p><span style='color: green'><b>This run has a full cache store.</b></span> Live runs have a much lower chance of reproducing correctly than an experiment script, but you can utilize this cache with:"
                f'<pre>manager = ArtifactManager("{manager.experiment_name}", cache_path="{store_entire_run_path}", dry_cache=True)</pre></p>'
            )
    else:
        html_lines.append(
            f"<p id='run-string'>Run string: <pre>{manager.run_line}</pre></p>"
        )

        if manager.store_entire_run:
            html_lines.append(
                f"<p><span style='color: green'><b>This run has a full cache store.</b></span> Reproduce with:"
                f"<pre>{manager.reproduction_line}</pre></p>"
            )

    if manager.notes is not None and manager.notes is not None and manager.notes != "":
        html_lines.append("<h3>Notes</h3>")
        notes = manager.notes
        notes = notes.replace("\n", "</br>")
        html_lines.append(f"<p>{notes}</p>")

    return html_lines


def render_report_toc() -> List[str]:
    """Render table of contents for the overall report."""
    return [
        "<a name='top'></a>" "<h2>Table of Contents</h2>",
        "<ul id='toc'>",
        "<li><a href='#reportables'>Reportables</a></li>"
        "<li><a href='#map'>Map</a></li>",
        "<li><a href='#stage_detail'>Stage Detail</a></li>",
        "<li><a href='#args'>Args</a></li>",
        "</ul>",
    ]


def render_report_reportables_toc(manager) -> List[str]:
    """Render the table of contents for the reportables."""
    html_lines = []

    # TODO: (01/27/2022) we can do this just by getting None group in grouped.
    # (will have to modify get_grouped_reportables to include None)
    non_grouped_reportables = manager.get_ungrouped_reportables()
    grouped_reportables = manager.get_grouped_reportables()

    html_lines.append("<a name='reportables'></a>")
    html_lines.append("<h2>Reportables</h2><ul id='reportables-list'>")
    for group in grouped_reportables:
        html_lines.append(f"<li>{group}<ul>")
        for reportable in grouped_reportables[group]:
            html_lines.append(
                f"<li><a href='#{reportable.name}'>{reportable.name}</a></li>"
            )
        html_lines.append("</ul></li>")
    for reportable in non_grouped_reportables:
        html_lines.append(
            f"<li><a href='#{reportable.name}'>{reportable.name}</a></li>"
        )
    html_lines.append("</ul>")
    return html_lines


def render_report_all_reportables(
    manager,
    reportables_path: str,
    override_display_path: str = None,
    notebook: bool = False,
) -> List[str]:
    """Get the HTML for displaying all reportables output."""
    html_lines = []

    non_grouped_reportables = manager.get_ungrouped_reportables()
    grouped_reportables = manager.get_grouped_reportables()

    # render grouped reportables
    for group in grouped_reportables:
        html_lines.append(f"<h3 class='reportable-group-title'>{group}</h3>")
        for reportable in grouped_reportables[group]:
            html_lines.extend(
                render_reportable(
                    reportable,
                    manager,
                    reportables_path,
                    override_display_path,
                    notebook,
                )
            )

    # render all non-grouped reportables
    for reportable in non_grouped_reportables:
        html_lines.extend(
            render_reportable(
                reportable, manager, reportables_path, override_display_path, notebook
            )
        )

    return html_lines


def render_report_stage_map(manager, graphs_path) -> List[str]:
    """Generate and write out the graphviz graph for the stages and return the html to display it."""
    graph = map_full_svg(manager)
    with open(f"{graphs_path}/full.gv", "w") as outfile:
        outfile.write(graph.source)

    html_lines = []
    html_lines.append("<a name='map'></a>")
    html_lines.append("<h2>Process/Stages Map</h2>")
    html_lines.append("<p><a href='#top'>back to top</a></p>")
    html_lines.append(render_graph(graph))
    return html_lines


def render_report_detailed_stage_maps(manager, graphs_path) -> List[str]:
    """Generate and write out a more detailed graphviz graph for each record and return the html."""
    html_lines = []
    html_lines.append("<a name='stage_detail'></a>")
    html_lines.append("<h2>Stage Data Detail</h2>")
    html_lines.append("<p><a href='#top'>back to top</a></p>")

    for index, record in enumerate(manager.records):
        graph = map_single_svg(manager, record)
        with open(f"{graphs_path}/record_{index}.gv", "w") as outfile:
            outfile.write(graph.source)
        html_lines.append(render_graph(graph))

    html_lines.append("<a name='args'></a>")
    html_lines.append("<h2>Args</h2>")
    html_lines.append("<p><a href='#top'>back to top</a></p>")
    return html_lines


def render_report_argset_dump(manager) -> List[str]:
    """Dump out the JSON for all args used by manager."""
    html_lines = []

    argsets = manager.get_all_argsets()
    for argset in argsets:
        name = argset.name
        args_hash = argset.hash
        html_lines.append(f"<h4>{name} - {args_hash}</h4>")
        html_lines.append("<pre>")

        def stringify(x):
            # make html-safe
            return str(x).replace("<", "&lt;").replace(">", "&gt;")

        argset_data = asdict(argset)
        del argset_data["name"]
        del argset_data["overwrite"]
        del argset_data["hash"]
        html_lines.append(json.dumps(argset_data, indent=2, default=stringify))
        html_lines.append("</pre>")

    return html_lines


def prepare_report_path(output_path, report_name):
    """Set up any necessary folders for a report at the given location. This will not error if the location already has a report in it, but will remove existing reportables and graphs."""

    folder_path = os.path.join(output_path, report_name)
    logging.info("Preparing report path '%s'..." % folder_path)

    graphs_path = os.path.join(folder_path, "graphs")
    reportables_path = os.path.join(folder_path, "reportables")

    if os.path.exists(graphs_path):
        shutil.rmtree(graphs_path)
    if os.path.exists(reportables_path):
        shutil.rmtree(reportables_path)

    if not os.path.exists(folder_path):
        os.makedirs(folder_path, exist_ok=True)
    os.mkdir(graphs_path)
    os.mkdir(reportables_path)

    return folder_path, graphs_path, reportables_path


def run_report(manager, output_path, name, css_path=None):
    """Generate a full HTML report for the given manager and passed argsets.

    Args:
        manager (ArtifactManager): The manager to base the graph off of.
        output_path (str): The string path to the root directory of where you want the report stored.
        name (str): The name to store this report under (will generate a folder of this name in the
            output_path.
        css_path (str): The path to a css file to use for styling the report. (This file will get
            copied into the output_path/name folder.)
    """

    # TODO: (01/26/2022) in principle this is where we could check manager if stored, and store if not.

    folder_path, graphs_path, reportables_path = prepare_report_path(output_path, name)

    html_lines = ["<html>"]
    html_lines.extend(render_report_head(manager))
    html_lines.append("<body>")
    html_lines.extend(render_report_info_block(manager))
    html_lines.extend(render_report_toc())

    # render reportables
    if len(manager.reportables) > 0:
        html_lines.extend(render_report_reportables_toc(manager))
        html_lines.extend(render_report_all_reportables(manager, reportables_path))

    # render graphs
    html_lines.extend(render_report_stage_map(manager, graphs_path))
    html_lines.extend(render_report_detailed_stage_maps(manager, graphs_path))

    # output the JSON args for each argset
    html_lines.extend(render_report_argset_dump(manager))
    html_lines.append("</body></html>")

    # write out all the things!
    with open(f"{folder_path}/index.html", "w") as outfile:
        outfile.writelines(html_lines)

    with open(f"{folder_path}/run_info.json", "w") as outfile:
        json.dump(manager.run_info, outfile, indent=4)

    if css_path is not None:
        if not os.path.exists(css_path):
            logging.warning("Reports CSS file %s not found" % css_path)
        else:
            shutil.copyfile(css_path, f"{folder_path}/style.css")

    return folder_path, graphs_path, reportables_path


def _add_record_subgraph(dot, record_index, record, manager, detailed=True):
    with dot.subgraph(name=f"cluster_{record_index}") as c:
        c.attr(color=str(_get_color(record_index)))
        c.attr(style="filled")
        if record.args is not None:
            c.attr(
                label="record "
                + str(record_index)
                + "\\nargs: "
                + str(record.args.name)
            )
        else:
            c.attr(label="record " + str(record_index) + "\\nargs: None")
        for stage in record.stages:
            stage_name = f"{record_index}_{stage}"
            c.node(stage_name, stage, style="filled", fillcolor="white", fontsize="12")

        for index, artifact in enumerate(manager.artifacts):
            if artifact.init_record == record:
                if detailed:
                    table = (
                        "<<table border='0' cellborder='1' cellspacing='0'><tr><td>"
                        + "<b>"
                        + str(artifact.name)
                        + "</b>"
                        + "</td></tr><tr><td width='0'>"
                        + artifact.html_safe()
                        + "</td></tr><tr><td>"
                        + artifact.file
                        + "</td></tr></table>>"
                    )
                    c.node(
                        f"a{index}",
                        table,
                        shape="none",
                        fontsize="8",
                        padding="0",
                        margin="0",
                        width="0",
                    )
                else:
                    c.node(
                        f"a{index}",
                        str(artifact.name),
                        shape="rectangle",
                        fontsize="10",
                        height=".20",
                    )

        for index, input_set in enumerate(record.stage_inputs):
            for input_index in input_set:
                if input_index != -1:
                    dot.edge(
                        f"a{input_index}",
                        f"{record_index}_{record.stages[index]}",
                        arrowsize=".65",
                    )
        for index, output_set in enumerate(record.stage_outputs):
            for output_index in output_set:
                dot.edge(
                    f"{record_index}_{record.stages[index]}",
                    f"a{output_index}",
                    arrowsize=".65",
                )


def update_report_index(experiments_path, reports_root_dir):
    """Generate a nice index.html with a summary line for each experiment run report in the
    passed directory.

    Args:
        experiments_path (str): The path to where the experiment scripts are, used to get any top
            experiment description comments from the file.
        reports_root_dir (str): The directory containing the report folders. This is where the output
            index.html is placed.
    """
    logging.info("Updating report index...")

    runs = []
    experiment_runs = {}

    no_info_runs = []
    no_index_runs = []
    informal_runs = (
        []
    )  # this is a combo of both no_info and no_index, they get listed together but distinguishing them to construct links differently is important

    # gather all reports
    for filename in os.listdir(reports_root_dir):
        if filename == "_latest":
            continue

        full_filename = f"{reports_root_dir}/{filename}"
        if not os.path.isdir(full_filename):
            continue

        if os.path.exists(f"{full_filename}/run_info.json"):
            with open(f"{full_filename}/run_info.json", "r") as infile:
                info = json.load(infile)

                info["order_timestamp"] = datetime.datetime.strptime(
                    info["timestamp"], utils.TIMESTAMP_FORMAT
                )

                runs.append(info)
                experiment_name = info["experiment_name"]
                if experiment_name not in experiment_runs:

                    # try to get comment on experiment
                    comment = ""
                    try:
                        with open(
                            f"{experiments_path}/{experiment_name}.py", "r"
                        ) as infile:
                            lines = infile.readlines()
                            comment = utils.get_py_opening_comment(lines)
                    except:  # noqa: E722 -- there's not really any need to handle this, either we get the comment or we don't.
                        pass

                    experiment_runs[experiment_name] = {"runs": [], "comment": comment}

                experiment_runs[experiment_name]["runs"].append(info)
        elif os.path.exists(f"{full_filename}/index.html"):
            informal_runs.append(filename)
            no_info_runs.append(filename)
        else:
            informal_runs.append(filename)
            no_index_runs.append(filename)

    logging.info("    %s labeled reports found", str(len(runs)))
    logging.info("    %s informal runs found", str(len(informal_runs)))

    html = [
        "<html>",
        "<head><title>Reports Index</title>",
        "<link rel='stylesheet' href='style.css'></head>",
        "<body>",
        "<h1 id='title'>Reports</h1>",
        "<p><a href='_latest/index.html'>View latest report</a></p>",
    ]

    # experiment_index =
    html.extend(["<h2>Experiments</h2>", "<ul>"])
    for experiment in experiment_runs:
        html.append(f"<li><a href='#{experiment}'>{experiment}</a></li>")
    html.append("<li><a href='#ALL'>ALL RUNS</a></li>")
    html.append("</ul>")

    for experiment in experiment_runs:
        html.extend(
            [
                f"<a name='{experiment}'></a>",
                f"<h2>{experiment}</h2>",
                f"<p>{experiment_runs[experiment]['comment']}</p>",
            ]
        )

        # order the runs
        ex_runs = sorted(
            experiment_runs[experiment]["runs"],
            key=lambda i: i["order_timestamp"],
            reverse=True,
        )

        html.append("<ul>")
        for run in ex_runs:
            html.append(_get_run_index_line(run))
        html.append("</ul>")

    html.extend(["<a name='ALL'></a>", "<h2>All labeled runs</h2>"])

    runs = sorted(runs, key=lambda i: i["order_timestamp"], reverse=True)
    html.append("<ul>")
    for run in runs:
        html.append(_get_run_index_line(run))
    html.append("</ul>")

    html.append("<h2>Informal runs</h2>")
    html.append("<ul>")
    for filename in informal_runs:
        if filename in no_info_runs:
            html.append(f"<li><a href='{filename}/index.html'>{filename}</a></li>")
        elif filename in no_index_runs:
            html.append(f"<li><a href='{filename}'>{filename}</a> (NO REPORT)</li>")
    html.append("</ul>")

    with open(f"{reports_root_dir}/index.html", "w") as outfile:
        outfile.writelines(html)


def map_single_svg(manager, record, detailed=True, colors=None):
    """Create a graphviz dot graph of the stages/artifacts for the given record.

    Args:
        manager (ArtifactManager): the manager to base the graph off of.
        record (Record): The specific record to make the graph for.
        detailed (bool): Whether to include artifact preview and caching information or not.
        colors: A list or dictionary of colors to use if you wish to override the default.

    Important:
        For this function to run successfully, graphviz must be installed.
    """
    _assign_colors(colors)

    dot = Digraph()
    dot.attr(compound="true")
    dot.attr(nodesep=".2")
    dot.attr(ranksep=".2")

    for index, record_check in enumerate(manager.records):
        if record_check == record:
            _add_record_subgraph(dot, index, record, manager, detailed)

    dot.format = "svg"
    return dot


def map_full_svg(manager, detailed=False, colors=None):
    """Create a graphviz dot graph for the entire experiment. (Maps out each stage and
    the input/output artifacts for each.)

    Args:
        manager (ArtifactManager): the manager to base the graph off of.
        detailed (bool): Whether to include artifact preview and caching information or not.
        colors: A list or dictionary of colors to use if you wish to override the default.

    Important:
        For this function to run successfully, graphviz must be installed.
    """
    _assign_colors(colors)

    dot = Digraph()
    dot.attr(compound="true")
    dot.attr(fontsize="10")
    dot.attr(nodesep=".15")
    dot.attr(ranksep=".15")
    # dot.attr(rankdir='LR')

    for index, record in enumerate(manager.records):
        _add_record_subgraph(dot, index, record, manager, detailed)

    # add record connections to any aggregate stages
    for index, record in enumerate(manager.records):
        for prev_record in record.input_records:
            if prev_record == record:
                continue
            # prev_record_index =
            for jndex, manager_record in enumerate(manager.records):
                if manager_record == prev_record:
                    dot.edge(
                        f"{jndex}_{manager.records[jndex].stages[-1]}",
                        f"{index}_{record.stages[0]}",
                        ltail=f"cluster_{jndex}",
                    )

    dot.format = "svg"
    return dot


def _assign_colors(colors=None):
    global COLORS, COLOR_VALS

    if colors is not None:
        if not isinstance(colors, list):
            colors = [colors]
        if isinstance(colors[0], tuple):
            COLORS = [color[0] for color in COLORS]
            COLOR_VALS = {color[0]: color[1] for color in COLORS}
        COLORS = colors


def _get_color(index):
    index = index % len(COLORS)
    name = COLORS[index]
    if name in COLOR_VALS:
        return COLOR_VALS[name]
    else:
        return name


def render_graph(graph):
    """Attempts to return the unicode text for the graph svg."""
    try:
        # return graph.pipe().decode('utf-8')
        return graph.pipe().decode("ascii")
    except ExecutableNotFound:
        logging.error(
            "Graphviz not installed, if using conda try 'conda install python-graphviz'."
        )
        return "<p style='color: red'>No graphviz executable found, cannot render stage maps.</p>"
    except Exception as e:
        logging.error("Graphviz error: %s", e)
        return f"<p style='color: red'>{e}</p>"


def render_reportable(
    reportable: Reportable,
    manager,
    reportables_path: str,
    override_display_path: str = None,
    notebook: bool = False,
) -> List[str]:
    """Render a reportable to file and get the HTML to display it."""
    html_lines = []

    # get associated record index
    color_string = "<span style='background-color: "
    for index, record in enumerate(manager.records):
        if record == reportable.record:
            color_string += _get_color(index) + "'>&nbsp;&nbsp;&nbsp;&nbsp;</span>"

    reportable.path = reportables_path  # path relative to experiment for saving
    reportable.render()
    if override_display_path is not None:
        reportable.path = override_display_path
    else:
        reportable.path = "reportables"  # path relative to html report for displaying
    if notebook:
        html_lines.append(
            "<div class='reportable' style='border: 1px solid gray; display: inline-block; vertical-align: top; padding: 5px; padding-top: 0px; padding-bottom: 0px; margin: 5px;'>"
        )
    else:
        html_lines.append("<div class='reportable'>")
    html_lines.append(f"<a name='{reportable.name}'></a>")
    html_lines.append(f"<h3>{reportable.name} {color_string}</h3>")
    reportable_html = reportable.html()
    if isinstance(reportable_html, list):
        html_lines.extend(reportable_html)
    else:
        html_lines.append(reportable_html)
    if not notebook:
        html_lines.append("<p><a href='#reportables'>back to reportables</a></p>")
    html_lines.append("</div>")

    return html_lines


def _get_run_index_line(run):
    desc_line = "<li>"

    if run["status"] == "complete":
        desc_line += "<span style='background-color: green'>&nbsp;&nbsp;</span>"
    elif run["status"] == "incomplete":
        desc_line += "<span style='background-color: orange'>&nbsp;&nbsp;</span>"
    elif run["status"] == "error":
        desc_line += "<span style='background-color: red'>&nbsp;&nbsp;</span>"
    elif run["status"] == "LIVE":
        desc_line += "<span style='background-color: cyan'>&nbsp;&nbsp;</span>"

    desc_line += f" <a href='{run['reference']}/index.html'>{run['reference']}</a> "

    if "hostname" in run:
        desc_line += f"[{run['hostname']}] "

    if run["status"] == "error":
        desc_line += f"<span style='color: red'>{run['error']}</span> "

    if run["full_store"]:
        desc_line += "<span style='color: magenta'><b>(full store)</b></span> "

    if "--dry-cache" in run["cli"] and "--cache" in run["cli"]:
        desc_line += "<span style='color: cyan'><b>(from stored)</b></span> "

    # params
    for params in run["args"]:
        args = [pair[0] for pair in run["args"][params]]
        desc_line += f"<b>{params}</b>({','.join(args)}) "

    # add notes if applicable
    if "notes" in run and run["notes"] is not None and run["notes"] != "":
        shortform_notes = run["notes"]
        if "\n" in shortform_notes:
            shortform_notes = shortform_notes[: shortform_notes.index("\n")]
        desc_line += f"</br><span style='color: #444; font-size: 10pt; font-style: italic;'>{shortform_notes}</span>"

    # add the runline
    desc_line += f"</br>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<span style='color: #a0a0a0; font-family: monospace'>{run['cli']}</span>"

    desc_line += "</li>"

    return desc_line



import sys

sys.path.append("test/examples")

from experiments import basic
from stages.basic_stages import get_data, sum_data
from params.basic import Args


from curifactory.manager import ArtifactManager
from curifactory.procedure import Procedure
from curifactory.utils import get_configuration


def test_import():
    """Importing the library should not throw errors!"""
    import curifactory  # noqa: F401 -- we're just trying to make sure it doesn't break


def test_ensure_config(configuration):
    """This is just to make sure the autouse fixture is doing what I expect."""
    assert get_configuration() == configuration


def test_experiment():
    manager = ArtifactManager(dry=True)
    basic.run(
        [
            Args(name="test1", starting_data=[1, 2, 3, 4]),
            Args(name="test2", starting_data=[4, 5, 6]),
        ],
        manager,
    )
    assert manager.records[0].state["sum"] == 10
    assert manager.records[1].state["sum"] == 15


def test_lone_procedure(configuration):
    proc = Procedure([get_data, sum_data], manager=ArtifactManager(dry=True))
    record = proc.run(Args(starting_data=[2, 3]))
    assert record.output == 5



from setuptools import setup
import re


def get_property(prop):
    result = re.search(
        r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop),
        open("curifactory/__init__.py").read(),
    )
    return result.group(1)


with open("README.md", "r", encoding="utf-8") as infile:
    long_description = infile.read()


setup(
    name="curifactory",
    version=get_property("__version__"),
    description="An experiment workflow and organization tool",
    keywords=["research", "experiment", "workflow"],
    long_description_content_type="text/markdown",
    long_description=long_description,
    author="Nathan Martindale, Jason Hite, Scott L. Stewart, Mark Adams",
    author_email="curifactory-help@ornl.gov",
    python_requires=">=3.8",
    url="https://github.com/ORNL/curifactory",
    project_urls={
        "Documentation": "https://ornl.github.io/curifactory/latest/index.html"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
    ],
    packages=["curifactory"],
    package_data={"curifactory": ["data/*", "data/.dockerignore"]},
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "experiment=curifactory.experiment:main",
            "curifactory=curifactory.project:main",
        ]
    },
    install_requires=[
        "numpy",
        "pandas",
        "ipynb-py-convert",
        "graphviz",
        "matplotlib",
        "psutil",
        "dataclasses",  # for compatibility with python 3.6
    ],
)
